from typing import Annotated

import grpc
from grpc._cython.cygrpc import _ServicerContext
from grpc.aio._typing import MetadataType
from pydantic import SkipValidation


class ContextWrapper:
    def __init__(self, context: _ServicerContext, trace_ctx: dict[str, str] | None = None) -> None:
        if trace_ctx is None:
            trace_ctx = {}
        self._context = context
        self._trace_ctx = trace_ctx

    def __getattr__(self, name: str):
        return self._context


class Context:
    def __init__(self, grpc_context: ContextWrapper) -> None:
        self.grpc_context = grpc_context

    @property
    def meta(self) -> dict[str, str]:
        metadata = {key: value for key, value in self.grpc_context._context.invocation_metadata()}
        metadata.update(self.trace_context)
        return metadata

    async def abort(self, code: grpc.StatusCode, details: str = "", trailing_metadata: MetadataType = ()) -> None:
        await self.grpc_context._context.abort(code, details, trailing_metadata)

    @property
    def trace_context(self) -> dict[str, str]:
        return self.grpc_context._trace_ctx



GRPCContext = Annotated[Context, SkipValidation()]




import asyncio
import logging
from typing import Any, AsyncIterator, Callable, Literal

import fast_depends
import grpc
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
from grpc._cython.cygrpc import _ServicerContext
from pydantic import ValidationError

from fastgrpcio._utils import pydantic_error_to_grpc
from fastgrpcio.context import GRPCContext, ContextWrapper
from fastgrpcio.middlewares import BaseMiddleware
from fastgrpcio.schemas import BaseGRPCSchema

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
logger = logging.getLogger(__name__)


class CreateHandlersMixins:
    _middlewares: list[BaseMiddleware]
    app_name: str
    app_package_name: str

    def _apply_middlewares(
        self,
        handler: Callable[..., Any],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        func_name: str,
    ) -> Any:
        if unary_type in ("Unary"):

            async def _apply_unary(request: Any, context: _ServicerContext) -> Any:
                context = ContextWrapper(context)
                async def call_next(req: Any, ctx: grpc.aio.ServicerContext) -> Any:
                    return await handler(req, ctx)

                for mw in reversed(self._middlewares):
                    prev_next = call_next

                    async def wrapper(
                        req: Any,
                        ctx: grpc.aio.ServicerContext,
                        mw: BaseMiddleware = mw,
                        nxt: Callable[..., Any] = prev_next,
                    ) -> Any:
                        return await mw.handle_unary(
                            request=req,
                            context=ctx,
                            call_next=nxt,
                            user_func=user_func,
                            request_model=request_model,
                            response_class=response_class,
                            handler=handler,
                            unary_type=unary_type,
                            app_name=self.app_name,
                            app_package_name=self.app_package_name,
                            func_name=func_name,
                        )

                    call_next = wrapper

                return await call_next(request, context)

            return _apply_unary

        elif unary_type in ("ServerStreaming", "BidiStreaming"):

            async def _apply_server_stream(request: Any, context: _ServicerContext) -> AsyncIterator[Any]:
                context = ContextWrapper(context)
                async def call_next(req: Any, ctx: _ServicerContext) -> AsyncIterator[Any]:
                    async for resp in handler(req, ctx):
                        yield resp

                for mw in reversed(self._middlewares):
                    prev_next = call_next

                    async def wrapper(
                        req: Any,
                        ctx: grpc.aio.ServicerContext,
                        mw: BaseMiddleware = mw,
                        nxt: Callable[..., Any] = prev_next,
                    ) -> AsyncIterator[Any]:
                        async for resp in mw.handle_stream(
                            request=req,
                            context=ctx,
                            call_next=nxt,
                            user_func=user_func,
                            request_model=request_model,
                            response_class=response_class,
                            handler=handler,
                            unary_type=unary_type,
                            app_name=self.app_name,
                            app_package_name=self.app_package_name,
                            func_name=func_name,
                        ):
                            yield resp

                    call_next = wrapper

                async for resp in call_next(request, context):
                    yield resp

            return _apply_server_stream

        elif unary_type in ("ClientStreaming"):

            async def _apply_client_stream(request: AsyncIterator[Any], context: _ServicerContext) -> Any:
                context = ContextWrapper(context)
                async def call_next(req_stream: AsyncIterator[Any], ctx: grpc.aio.ServicerContext) -> Any:
                    return await handler(req_stream, ctx)

                for mw in reversed(self._middlewares):
                    prev_next = call_next

                    async def wrapper(
                        req_stream: AsyncIterator[Any],
                        ctx: grpc.aio.ServicerContext,
                        mw: BaseMiddleware = mw,
                        nxt: Callable[..., Any] = prev_next,
                    ) -> Any:
                        return await mw.handle_client_stream(
                            request=req_stream,
                            context=ctx,
                            call_next=nxt,
                            user_func=user_func,
                            request_model=request_model,
                            response_class=response_class,
                            handler=handler,
                            unary_type=unary_type,
                            app_name=self.app_name,
                            app_package_name=self.app_package_name,
                            func_name=func_name,
                        )

                    call_next = wrapper

                return await call_next(request, context)

            return _apply_client_stream

    def _make_unary_handler(
        self,
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        func_name: str,
    ) -> Callable[..., Any]:
        async def handler(request_proto: Message, context: ContextWrapper) -> Any:
            request_dict: dict[str, Any] = MessageToDict(request_proto)
            try:
                pydantic_request = request_model.model_validate(request_dict)
            except ValidationError as e:
                grpc_status_obj = pydantic_error_to_grpc(e)
                await context._context.abort_with_status(grpc_status_obj)
                return

            injected = fast_depends.inject(user_func)

            grpc_context = GRPCContext(context)
            result = (
                await injected(pydantic_request, context=grpc_context)
                if asyncio.iscoroutinefunction(injected)
                else injected(pydantic_request, context=grpc_context)
            )

            if isinstance(result, response_class):
                return result

            return response_class(**result.model_dump())

        handler = self._apply_middlewares(handler, user_func, request_model, response_class, unary_type="Unary", func_name=func_name)
        return handler

    def _make_server_stream_handler(
        self,
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        func_name: str,
    ) -> Callable[..., Any]:
        async def handler(request_proto: Message, context: ContextWrapper) -> AsyncIterator[Any]:
            request_dict: dict[str, Any] = MessageToDict(request_proto)
            try:
                pydantic_request = request_model.model_validate(request_dict)
            except ValidationError as e:
                grpc_status_obj = pydantic_error_to_grpc(e)
                await context._context.abort_with_status(grpc_status_obj)
                return

            injected = fast_depends.inject(user_func)
            grpc_context = GRPCContext(context)
            result = injected(pydantic_request, context=grpc_context)
            if asyncio.iscoroutine(result):
                result = await result

            async for item in result:
                yield response_class(**item.model_dump())

        handler = self._apply_middlewares(
            handler, user_func, request_model, response_class, unary_type="ServerStreaming", func_name=func_name
        )
        return handler

    def _make_client_stream_handler(
        self,
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        func_name: str,
    ) -> Callable[..., Any]:
        async def handler(request_iterator: AsyncIterator[Message], context: ContextWrapper) -> Any:
            async def pydantic_request_gen() -> AsyncIterator[Any]:
                async for msg in request_iterator:
                    msg_dict: dict[str, Any] = MessageToDict(msg)
                    try:
                        yield request_model.model_validate(msg_dict)
                    except ValidationError as e:
                        grpc_status_obj = pydantic_error_to_grpc(e)
                        await context._context.abort_with_status(grpc_status_obj)
                        return

            injected = fast_depends.inject(user_func)
            grpc_context = GRPCContext(context)
            result = (
                await injected(pydantic_request_gen(), context=grpc_context)
                if asyncio.iscoroutinefunction(user_func)
                else injected(pydantic_request_gen(), context=grpc_context)
            )

            if isinstance(result, response_class):
                return result
            return response_class(**result.model_dump())

        handler = self._apply_middlewares(
            handler, user_func, request_model, response_class, unary_type="ClientStreaming", func_name=func_name
        )
        return handler

    def _make_bidi_stream_handler(
        self,
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        func_name: str,
    ) -> Callable[..., Any]:
        async def handler(request_iterator: AsyncIterator[Message], context: ContextWrapper) -> AsyncIterator[Any]:
            async def pydantic_request_gen() -> AsyncIterator[Any]:
                async for msg in request_iterator:
                    msg_dict: dict[str, Any] = MessageToDict(msg)
                    try:
                        yield request_model.model_validate(msg_dict)
                    except ValidationError as e:
                        grpc_status_obj = pydantic_error_to_grpc(e)
                        await context._context.abort_with_status(grpc_status_obj)
                        return

            injected = fast_depends.inject(user_func)
            grpc_context = GRPCContext(context)

            result = injected(pydantic_request_gen(), context=grpc_context)
            if asyncio.iscoroutine(result):
                result = await result

            async for resp in result:
                yield response_class(**resp.model_dump())

        handler = self._apply_middlewares(handler, user_func, request_model, response_class, unary_type="BidiStreaming", func_name=func_name)
        return handler



from __future__ import annotations

from typing import Any, AsyncIterator, Awaitable, Callable, Literal
from google.protobuf.message import Message
try:
    from opentelemetry import trace
    from opentelemetry.propagate import extract, inject
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.trace import INVALID_SPAN, Span
except ImportError:
    raise ImportError(
        "opentelemetry-sdk is required to use TracingMiddleware. "
        "Please install it with 'pip install fastgrpcio[otel]'"
    )

from fastgrpcio.context import ContextWrapper
from fastgrpcio.middlewares import BaseMiddleware
from fastgrpcio.schemas import BaseGRPCSchema


class TracingMiddleware(BaseMiddleware):

    def __init__(self, tracer_provider: TracerProvider) -> None:
        self.tracer_provider = tracer_provider

    def _start_root_if_needed(self, ctx: trace.Context, tracer: trace.Tracer) -> tuple[trace.Context, dict[str, str]]:
        current_span: Span = trace.get_current_span(context=ctx)
        metadata: dict[str, str] = {}

        if current_span.get_span_context() == INVALID_SPAN.get_span_context():
            with tracer.start_as_current_span("RootTrace") as root_span:
                root_ctx = trace.set_span_in_context(root_span)
                inject(metadata, context=root_ctx)
                return root_ctx, metadata
        return ctx, metadata

    @staticmethod
    def _set_rpc_attributes(
        span: Span,
        func_name: str,
        app_name: str,
        app_package_name: str,
    ) -> None:
        span.set_attribute("rpc.method", func_name)
        span.set_attribute("rpc.service", app_name)
        span.set_attribute("rpc.package", app_package_name)

    async def handle_unary(
        self,
        request: Message,
        context: ContextWrapper,
        call_next: Callable[[Any, ContextWrapper], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        app_name: str = "",
        app_package_name: str = "",
        func_name: str = "",
    ) -> Any:
        carrier: dict[str, str] = dict(context._context.invocation_metadata())
        ctx = extract(carrier)
        tracer = self.tracer_provider.get_tracer(__name__)
        ctx, _ = self._start_root_if_needed(ctx, tracer)

        with tracer.start_as_current_span(f"{app_package_name}/{app_name}/{func_name}", context=ctx) as span:
            self._set_rpc_attributes(span, func_name, app_name, app_package_name)
            inject(context._trace_ctx, context=trace.set_span_in_context(span))
            return await call_next(request, context)

    async def handle_client_stream(
        self,
        request: AsyncIterator[Message],
        context: ContextWrapper,
        call_next: Callable[[AsyncIterator[Any], ContextWrapper], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        app_name: str = "",
        app_package_name: str = "",
        func_name: str = "",
    ) -> Any:
        carrier = dict(context._context.invocation_metadata())
        ctx = extract(carrier)
        tracer = self.tracer_provider.get_tracer(__name__)
        ctx, _ = self._start_root_if_needed(ctx, tracer)

        with tracer.start_as_current_span(f"{app_package_name}/{app_name}/{func_name}", context=ctx) as span:
            self._set_rpc_attributes(span, func_name, app_name, app_package_name)
            inject(context._trace_ctx, context=trace.set_span_in_context(span))

            async def wrapped_stream() -> AsyncIterator[Any]:
                async for msg in request:
                    yield msg

            return await call_next(wrapped_stream(), context)

    async def handle_stream(
        self,
        request: Message,
        context: ContextWrapper,
        call_next: Callable[[Any, ContextWrapper], AsyncIterator[Message]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        app_name: str = "",
        app_package_name: str = "",
        func_name: str = "",
    ) -> AsyncIterator[Message]:
        carrier = dict(context._context.invocation_metadata())
        ctx = extract(carrier)
        tracer = self.tracer_provider.get_tracer(__name__)
        ctx, _ = self._start_root_if_needed(ctx, tracer)

        with tracer.start_as_current_span(f"{app_package_name}/{app_name}/{func_name}", context=ctx) as span:
            self._set_rpc_attributes(span, func_name, app_name, app_package_name)
            inject(context._trace_ctx, context=trace.set_span_in_context(span))

            async for response in call_next(request, context):
                yield response



from __future__ import annotations

import asyncio
import grpc
from typing import Any, AsyncIterator, Type, Callable

from google.protobuf import descriptor_pb2, descriptor_pool
from google.protobuf.message_factory import GetMessageClass
from google.protobuf.json_format import ParseDict, MessageToDict
from grpc_reflection.v1alpha import reflection_pb2, reflection_pb2_grpc

try:
    from opentelemetry import trace
    from opentelemetry.propagate import extract, inject
    from opentelemetry.trace import INVALID_SPAN
    _OTEL_ENABLED = True
except ImportError:
    _OTEL_ENABLED = False

    class DummyTracer:
        def start_as_current_span(self, name: str, context: Any = None):
            class DummySpan:
                def __enter__(self): return None
                def __exit__(self, *a): return False
            return DummySpan()

    class DummyTrace:
        @staticmethod
        def get_tracer(name: str) -> DummyTracer: return DummyTracer()

    class DummyPropagate:
        @staticmethod
        def extract(carrier: dict[str, str]) -> Any: return None
        @staticmethod
        def inject(carrier: dict[str, str], context: Any = None) -> None: return None

    trace = DummyTrace()
    extract = DummyPropagate.extract
    inject = DummyPropagate.inject
    INVALID_SPAN = object()


class GRPCClient:
    def __init__(
        self,
        target: str,
        use_tls: bool = False,
        *,
        max_retries: int = 3,
        retry_backoff: float = 0.5,
        retry_exceptions: tuple[type[BaseException], ...] = (
            grpc.aio.AioRpcError,
            ConnectionError,
            TimeoutError,
        ),
    ) -> None:
        self.target = target
        self.use_tls = use_tls
        self.channel: grpc.aio.Channel | None = None
        self.tracer = trace.get_tracer(__name__)
        self.max_retries = max_retries
        self.retry_backoff = retry_backoff
        self.retry_exceptions = retry_exceptions

    async def __aenter__(self) -> GRPCClient:
        creds = grpc.ssl_channel_credentials() if self.use_tls else None
        self.channel = (
            grpc.aio.secure_channel(self.target, creds)
            if creds else grpc.aio.insecure_channel(self.target)
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.channel:
            await self.channel.close()

    async def _get_service_descriptor(
        self,
        service_name: str,
    ) -> tuple[Any, descriptor_pool.DescriptorPool]:
        if not self.channel:
            raise RuntimeError("Channel is not initialized")

        stub = reflection_pb2_grpc.ServerReflectionStub(self.channel)

        list_req = reflection_pb2.ServerReflectionRequest(list_services="")
        call = stub.ServerReflectionInfo()
        await call.write(list_req)
        await call.done_writing()
        list_response = await call.read()

        services = [s.name for s in list_response.list_services_response.service]
        if service_name not in services:
            raise ValueError(f"Service '{service_name}' not found. Found: {services}")

        file_req = reflection_pb2.ServerReflectionRequest(file_containing_symbol=service_name)
        call = stub.ServerReflectionInfo()
        await call.write(file_req)
        await call.done_writing()
        file_response = await call.read()

        file_proto = file_response.file_descriptor_response.file_descriptor_proto[0]
        file_desc_proto = descriptor_pb2.FileDescriptorProto.FromString(file_proto)

        pool = descriptor_pool.DescriptorPool()
        pool.Add(file_desc_proto)
        service_desc = pool.FindServiceByName(service_name)

        return service_desc, pool

    def _create_messages(
        self,
        pool: descriptor_pool.DescriptorPool,
        method_desc: Any,
    ) -> tuple[Type, Type]:
        request_cls = GetMessageClass(pool.FindMessageTypeByName(method_desc.input_type.full_name))
        response_cls = GetMessageClass(pool.FindMessageTypeByName(method_desc.output_type.full_name))
        return request_cls, response_cls

    async def _prepare_tracing_context(
        self,
        metadata: dict[str, str] | None = None,
    ) -> tuple[dict[str, str], Any]:
        if not _OTEL_ENABLED:
            return dict(metadata or []), None

        ctx = extract(dict(metadata or []))
        current_span = trace.get_current_span(context=ctx)
        try:
            invalid_ctx = INVALID_SPAN.get_span_context()
        except AttributeError:
            invalid_ctx = None

        if hasattr(current_span, "get_span_context") and current_span.get_span_context() == invalid_ctx:
            metadata_dict: dict[str, str] = {}
            with self.tracer.start_as_current_span("RootTrace") as span:
                ctx = trace.set_span_in_context(span)
                inject(metadata_dict, context=ctx)
            return metadata_dict, ctx

        metadata_dict = dict(metadata or [])
        inject(metadata_dict, context=ctx)
        return metadata_dict, ctx

    async def _retry_call(self, func: Callable[[], Any]) -> Any:
        last_exc: Exception | None = None
        for attempt in range(1, self.max_retries + 1):
            try:
                return await func()
            except self.retry_exceptions as exc:
                last_exc = exc
                if attempt == self.max_retries:
                    raise
                delay = self.retry_backoff * (2 ** (attempt - 1))
                await asyncio.sleep(delay)
        if last_exc:
            raise last_exc


    async def unary_unary(
        self,
        service_name: str,
        method_name: str,
        body: dict[str, Any],
        *,
        metadata: list[tuple[str, str]] | None = None,
        timeout: float | None = 10,
    ) -> dict[str, Any]:
        if not self.channel:
            raise RuntimeError("Channel is not initialized")

        service_desc, pool = await self._get_service_descriptor(service_name)
        method_desc = service_desc.FindMethodByName(method_name)
        request_cls, response_cls = self._create_messages(pool, method_desc)
        request_msg = request_cls()
        ParseDict(body, request_msg)

        metadata_dict, ctx = await self._prepare_tracing_context(metadata)
        method_path = f"/{service_name}/{method_name}"

        call = self.channel.unary_unary(
            method_path,
            request_serializer=request_msg.SerializeToString,
            response_deserializer=response_cls.FromString,
        )

        async def do_call() -> dict[str, Any]:
            with self.tracer.start_as_current_span(f"grpc.unary_unary.{method_name}", context=ctx):
                response = await call(request_msg, metadata=metadata_dict.items(), timeout=timeout)
                return MessageToDict(response, preserving_proto_field_name=True)

        return await self._retry_call(do_call)

    async def unary_stream(
        self,
        service_name: str,
        method_name: str,
        body: dict[str, Any],
        *,
        metadata: list[tuple[str, str]] | None = None,
        timeout: float | None = 10,
    ) -> AsyncIterator[dict[str, Any]]:
        if not self.channel:
            raise RuntimeError("Channel is not initialized")

        service_desc, pool = await self._get_service_descriptor(service_name)
        method_desc = service_desc.FindMethodByName(method_name)
        request_cls, response_cls = self._create_messages(pool, method_desc)
        request_msg = request_cls()
        ParseDict(body, request_msg)

        metadata_dict, ctx = await self._prepare_tracing_context(metadata)
        method_path = f"/{service_name}/{method_name}"

        call = self.channel.unary_stream(
            method_path,
            request_serializer=request_msg.SerializeToString,
            response_deserializer=response_cls.FromString,
        )

        async def stream_call() -> AsyncIterator[dict[str, Any]]:
            with self.tracer.start_as_current_span(f"grpc.unary_stream.{method_name}", context=ctx):
                async for response in call(request_msg, metadata=metadata_dict.items(), timeout=timeout):
                    yield MessageToDict(response, preserving_proto_field_name=True)

        for attempt in range(1, self.max_retries + 1):
            try:
                async for item in stream_call():
                    yield item
                break
            except self.retry_exceptions as exc:
                if attempt == self.max_retries:
                    raise
                await asyncio.sleep(self.retry_backoff * (2 ** (attempt - 1)))

    async def stream_unary(
        self,
        service_name: str,
        method_name: str,
        body_stream: AsyncIterator[dict[str, Any]],
        *,
        metadata: list[tuple[str, str]] | None = None,
        timeout: float | None = 10,
    ) -> dict[str, Any]:
        if not self.channel:
            raise RuntimeError("Channel is not initialized")

        service_desc, pool = await self._get_service_descriptor(service_name)
        method_desc = service_desc.FindMethodByName(method_name)
        request_cls, response_cls = self._create_messages(pool, method_desc)

        async def req_iter() -> AsyncIterator[Any]:
            async for item in body_stream:
                msg = request_cls()
                ParseDict(item, msg)
                yield msg

        metadata_dict, ctx = await self._prepare_tracing_context(metadata)
        method_path = f"/{service_name}/{method_name}"

        call = self.channel.stream_unary(
            method_path,
            request_serializer=lambda m: m.SerializeToString(),
            response_deserializer=response_cls.FromString,
        )

        async def do_call() -> dict[str, Any]:
            with self.tracer.start_as_current_span(f"grpc.stream_unary.{method_name}", context=ctx):
                response = await call(req_iter(), metadata=metadata_dict.items(), timeout=timeout)
                return MessageToDict(response, preserving_proto_field_name=True)

        return await self._retry_call(do_call)

    async def stream_stream(
        self,
        service_name: str,
        method_name: str,
        body_stream: AsyncIterator[dict[str, Any]],
        *,
        metadata: list[tuple[str, str]] | None = None,
        timeout: float | None = 10,
    ) -> AsyncIterator[dict[str, Any]]:
        if not self.channel:
            raise RuntimeError("Channel is not initialized")

        service_desc, pool = await self._get_service_descriptor(service_name)
        method_desc = service_desc.FindMethodByName(method_name)
        request_cls, response_cls = self._create_messages(pool, method_desc)

        async def req_iter() -> AsyncIterator[Any]:
            async for item in body_stream:
                msg = request_cls()
                ParseDict(item, msg)
                yield msg

        metadata_dict, ctx = await self._prepare_tracing_context(metadata)
        method_path = f"/{service_name}/{method_name}"

        call = self.channel.stream_stream(
            method_path,
            request_serializer=lambda m: m.SerializeToString(),
            response_deserializer=response_cls.FromString,
        )

        for attempt in range(1, self.max_retries + 1):
            try:
                with self.tracer.start_as_current_span(f"grpc.stream_stream.{method_name}", context=ctx):
                    async for response in call(req_iter(), metadata=metadata_dict.items(), timeout=timeout):
                        yield MessageToDict(response, preserving_proto_field_name=True)
                break
            except self.retry_exceptions:
                if attempt == self.max_retries:
                    raise
                await asyncio.sleep(self.retry_backoff * (2 ** (attempt - 1)))



import logging
from typing import Any, AsyncIterator, Awaitable, Callable, Literal

import grpc
from google.protobuf.message import Message

from fastgrpcio.schemas import BaseGRPCSchema

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
logger = logging.getLogger(__name__)


class BaseMiddleware:
    async def handle_unary(
        self,
        request: Message,
        context: grpc.aio.ServicerContext,
        call_next: Callable[[Any, grpc.aio.ServicerContext], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> Any:
        response = await call_next(request, context)
        return response

    async def handle_stream(
        self,
        request: Message,
        context: grpc.aio.ServicerContext,
        call_next: Callable[..., Any],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> AsyncIterator[Message]:
        async for resp in call_next(request, context):
            yield resp

    async def handle_client_stream(
        self,
        request: AsyncIterator[Message],
        context: grpc.aio.ServicerContext,
        call_next: Callable[[Any, grpc.aio.ServicerContext], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> Any:
        async def wrapped_stream() -> AsyncIterator[Any]:
            async for msg in request:
                yield msg

        response = await call_next(wrapped_stream(), context)
        return response


class LoggingMiddleware(BaseMiddleware):
    async def handle_unary(
        self,
        request: Message,
        context: grpc.aio.ServicerContext,
        call_next: Callable[[Any, grpc.aio.ServicerContext], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> Any:
        logger.info(f"[{unary_type}] - {user_func.__name__} - Received request")
        response = await call_next(request, context)
        logger.info(f"[{unary_type}] - {user_func.__name__} - Processed response")
        return response

    async def handle_stream(
        self,
        request: Message,
        context: grpc.aio.ServicerContext,
        call_next: Callable[..., Any],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> Any:
        logger.info(f"[{unary_type}] - {user_func.__name__} - Started streaming")
        async for resp in call_next(request, context):
            logger.info(f"[{unary_type}] - {user_func.__name__} - Streamed response chunk")
            yield resp
        logger.info(f"[{unary_type}] - {user_func.__name__} - Completed streaming")

    async def handle_client_stream(
        self,
        request: AsyncIterator[Message],
        context: grpc.aio.ServicerContext,
        call_next: Callable[[AsyncIterator[Any], grpc.aio.ServicerContext], Awaitable[Any]],
        user_func: Callable[..., Any],
        request_model: type[BaseGRPCSchema],
        response_class: type[BaseGRPCSchema],
        handler: Callable[..., Any],
        unary_type: Literal["Unary", "ServerStreaming", "ClientStreaming", "BidiStreaming"],
        **kwargs
    ) -> Any:
        async def wrapped_stream() -> AsyncIterator[Any]:
            async for msg in request:
                logger.info(f"[{unary_type}] - {user_func.__name__} - Recieved client message")
                yield msg

        logger.info(f"[{unary_type}] - {user_func.__name__} - Started receiving client stream")
        response = await call_next(wrapped_stream(), context)
        logger.info(f"[{unary_type}] - {user_func.__name__} - Client stream completed")
        return response
